#!/usr/bin/python
import sys
import os
import unittest
import logging
import pyuppaal
from opaal.model_parsers.pyuppaal import util
from opaal.model_parsers.pyuppaal.gensuccgen import GenerateSuccessorGenerator
from opaal.model_parsers.pyuppaal.simplify import SimplifyModel

class TestPYUPPAALGenSuccGen(unittest.TestCase):
    def setUp(self):
        fname = self._testMethodName[len("test_"):]
        try:
            self.file = open(os.path.join(os.path.dirname(__file__), 'gensuccgen', fname + '.xml'))
            if self.file:
                self.nta = pyuppaal.NTA.from_xml(self.file)
        except:
            self.file = None
            self.nta = None

    def test_simple1(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.gx, 0)
        self.assertEqual(s1.gy, 0)
        self.assertEqual(s1.Process_lx, 0)
        self.assertEqual(s1.Process_ly, 0)
        self.assertEqual(s1.Process, 1)

        succs = list(s.successors(s1))

        self.assertEqual(len(succs), 1)
        s1 = succs[0]
        self.assertEqual(s1.gx, 0)
        self.assertEqual(s1.gy, 0)
        self.assertEqual(s1.Process_lx, 0)
        self.assertEqual(s1.Process_ly, 0)
        self.assertEqual(s1.Process, 0)

    def test_simple_covered_by(self): #get the initial state and compare the federation against itself
        self.file = open(os.path.join(os.path.dirname(__file__), 'gensuccgen', 'simple_clocktrans1.xml'))
        self.nta = pyuppaal.NTA.from_xml(self.file)
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        self.assertTrue(s._covered_by_short(s1, s1))

    def test_covered_by(self): #get two states and compare them
        s = GenerateSuccessorGenerator(self.nta)

        s1 = s.get_initialstate()
        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 2)

        self.assertTrue(s._covered_by_short(succs[0], succs[1]))
        self.assertTrue(s._covered_by_short(succs[1], succs[0]))

    def test_sync_twoway(self):
        self.file = open(os.path.join(os.path.dirname(__file__), 'sync_twoway.xml'))
        self.nta = pyuppaal.NTA.from_xml(self.file)
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.A, 0)
        self.assertEqual(s1.B, 0)

    def test_sync_twoway_twochoices(self):
        self.file = open(os.path.join(os.path.dirname(__file__), 'sync_twoway_twochoices.xml'))
        self.nta = pyuppaal.NTA.from_xml(self.file)
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.A, 1)
        self.assertEqual(s1.B, 1)

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s1 = succs[0]

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 2)
        s1 = succs[0]
        self.assertEqual(s1.A, 0)
        s2 = succs[1]
        self.assertEqual(s2.A, 0)
        self.assertNotEqual(s1.B, s2.B)


    def test_simple_constraints(self): #simple test of complex contraints to test it still compiles
        s = GenerateSuccessorGenerator(self.nta)

        s1 = s.get_initialstate()

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 2)
        print succs[0], succs[1]

        self.assertTrue(succs[0].covers(succs[1]))
        self.assertTrue(succs[1].covers(succs[0]))

    def test_simple_clocktrans1(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.Process, 1)
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]

        self.assertEqual(s2.Process, 0)
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        #with max constant extrapolation
        self.assertEqual(str(s2fed), 'true')
        #before:
        #self.assertEqual(str(s2fed), '(5<=c & c==y)')

    def test_maxconst1(self):
        self.file = open(os.path.join(os.path.dirname(__file__), 'maxconst1.xml'))
        self.nta = pyuppaal.NTA.from_xml(self.file)
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()

        # initial state
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==y)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x==y)')

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 2)
        s3 = succs2[0] ; s4 = succs2[1]
        #assert about dbm
        #assume this is the 'c' location
        s3fed = s3._get_pyfed_from_state()
        self.assertEqual(str(s3fed), '(y<=100)')
        s4fed = s4._get_pyfed_from_state()
        self.assertEqual(str(s4fed), 'true')

    def test_maxconst5(self):
        self.file = open(os.path.join(os.path.dirname(__file__), 'maxconst5.xml'))
        self.nta = pyuppaal.NTA.from_xml(self.file)
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = s.successors(s1)
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), 'true')

        succs2 = s.successors(s2)
        self.assertEqual(len(succs2), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), 'true')

    def test_sync_bcast(self):
        self.file = open(os.path.join(os.path.dirname(__file__), 'sync_bcast.xml'))
        self.nta = pyuppaal.NTA.from_xml(self.file)

        s = GenerateSuccessorGenerator(self.nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        #B can do independent, or A can do action
        self.assertEqual(len(succs), 2)

        #early sync branch
        earlysync_succ = [succ for succ in succs if succ.A == 1][0]
        succs2 = list(s.successors(earlysync_succ))
        self.assertEqual(len(succs2), 1)
        succs3 = list(s.successors(succs2[0]))
        self.assertEqual(len(succs3), 0)
        #B has not synced
        self.assertEqual(1, succs2[0].B)

        #late sync branch
        latesync_succ = [succ for succ in succs if succ.A == 0][0]
        succs2 = list(s.successors(latesync_succ))
        self.assertEqual(len(succs2), 1)
        succs3 = list(s.successors(succs2[0]))
        self.assertEqual(len(succs3), 0)
        #B has synced
        self.assertEqual(2, succs2[0].B)

    def test_sync_bcast_many(self):
        self.file = open(os.path.join(os.path.dirname(__file__), 'sync_bcast_many.xml'))
        self.nta = pyuppaal.NTA.from_xml(self.file)
        succgen = GenerateSuccessorGenerator(self.nta)
        
        #check that all participate
        init_state = succgen.get_initialstate()
        self.assertEqual(init_state.mysum, 0)

        succs = list(succgen.successors(init_state))
        #all must participate
        self.assertEqual(len(succs), 1)
        succ = succs[0]

        self.assertEqual(succ.mysum, 15)

    def test_sync_simple_nop(self):
        succgen = GenerateSuccessorGenerator(self.nta)
        
        init_state = succgen.get_initialstate()
        self.assertEqual(init_state.x, 0)
        self.assertEqual(init_state.y, 0)

        succs = list(succgen.successors(init_state))
        self.assertEqual(len(succs), 1)
        succ = succs[0]

        self.assertEqual(succ.x, 0)
        self.assertEqual(succ.y, 0)

    def test_priorities1(self):
        #check simplify leaves the priorities alone
        simple_nta = SimplifyModel(self.nta).simplify()
        self.assertEqual(simple_nta.system, "system B < A;")

        s = GenerateSuccessorGenerator(self.nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        #only A can do action
        self.assertEqual(len(succs), 1)
        succ = succs[0]
        print succ, initial_state
        self.assertNotEqual(succ.A, initial_state.A)

        succs = list(s.successors(succ))
        #only B can do action
        self.assertEqual(len(succs), 1)
        succ = succs[0]
        self.assertNotEqual(succ.A, initial_state.A)
        self.assertNotEqual(succ.B, initial_state.B)

    def test_priorities2(self):
        #check simplify leaves the priorities alone
        simple_nta = SimplifyModel(self.nta).simplify()
        self.assertEqual(simple_nta.system, "system B < A;")
        
        s = GenerateSuccessorGenerator(self.nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        #only A can do action
        self.assertEqual(len(succs), 1)
        succ = succs[0]
        print succ, initial_state
        self.assertNotEqual(succ.A, initial_state.A)

        succs = list(s.successors(succ))
        #deadlock
        self.assertEqual(len(succs), 0)

    def test_invariant_delay(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")

        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(c<=4)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(c==4)')

    def test_upc_guard_true(self):
        s = GenerateSuccessorGenerator(self.nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 2)

        succs2a = list(s.successors(succs[0]))
        self.assertEqual(len(succs2a), 0)
        succs2b = list(s.successors(succs[1]))
        self.assertEqual(len(succs2b), 0)

    def test_upc_func(self):
        s = GenerateSuccessorGenerator(self.nta)

        s1 = s.get_initialstate()
        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 2)

    def test_upc_loop_while(self):
        s = GenerateSuccessorGenerator(self.nta)

        s1 = s.get_initialstate()
        s2 = list(s.successors(s1))[0]
        succs = list(s.successors(s2))
        self.assertEqual(len(succs), 1)

    def test_upc_loop_do_while(self):
        file = open(os.path.join(os.path.dirname(__file__), 'gensuccgen', 'upc_loop_do_while.xml'))
        self.nta = pyuppaal.NTA.from_xml(file)
        s = GenerateSuccessorGenerator(self.nta)

        s1 = s.get_initialstate()
        s2 = list(s.successors(s1))[0]
        succs = list(s.successors(s2))
        self.assertEqual(len(succs), 1)

    def test_upc_loop_for(self):
        s = GenerateSuccessorGenerator(self.nta)

        s1 = s.get_initialstate()
        s2 = list(s.successors(s1))[0]
        succs = list(s.successors(s2))
        self.assertEqual(len(succs), 1)

    def test_upc_update_func(self):
        s = GenerateSuccessorGenerator(self.nta)

        s1 = s.get_initialstate()
        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)

    def test_upc_update_array(self):
        s = GenerateSuccessorGenerator(self.nta)

        s1 = s.get_initialstate()
        succs = list(s.successors(s1))
        succs = list(s.successors(succs[0]))
        self.assertEqual(len(succs), 1)

    def test_clock_guard_then_reset(self):
        s = GenerateSuccessorGenerator(self.nta)
        initial_state = s.get_initialstate()

        print initial_state

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 1)
        print succs[0]

        succs2 = list(s.successors(succs[0]))

        self.assertEqual(len(succs2), 0)

    def test_clock_guard_then_reset2(self):
        s = GenerateSuccessorGenerator(self.nta)
        initial_state = s.get_initialstate()

        print initial_state

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 1)
        print succs[0]

        succs2 = list(s.successors(succs[0]))

        self.assertEqual(len(succs2), 1)
        print succs2[0]

    def test_clock_compare_two(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()

        # initial state
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x<42 & x==y)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x<42 & y-x<=-2)')

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 1)
        s3 = succs2[0]
        #assert about dbm
        s3fed = s3._get_pyfed_from_state()
        self.assertEqual(str(s3fed), '(x<42 & y-x<=-2)')

    def test_clock_equal_guard(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()

        # initial state
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(1<=x & x<=42)')

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 0)

    def test_committed_location(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        succs = list(s.successors(s1))
        #only one succ, committed
        self.assertEqual(len(succs), 1)
        s2 = succs[0]

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 1)
        s3 = succs2[0]

    def test_priorities_commited_locs(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        #print s1
        #self.assertFalse(s._get_object_file().state_is_committed(pointer(s1)))

        succs = list(s.successors(s1))
        #only one succ, synchronize
        self.assertEqual(len(succs), 1)
        s2 = succs[0]

        #two committed locs, but one has higher priority
        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 1)
        s3 = succs2[0]

        #remaining committed loc
        succs3 = list(s.successors(s3))
        self.assertEqual(len(succs3), 1)
        s4 = succs3[0]

        #deadlock
        succs4 = list(s.successors(s4))
        self.assertEqual(len(succs4), 0)

    def test_committed_urgency(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

        succs = list(s.successors(s1))
        #should be deadlock, time not allowed to elapse in initial state
        self.assertEqual(len(succs), 0)

    def test_committed_urgency2(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x==0)')

        #should be deadlock, time not allowed to elapse
        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 0)

    def test_committed_urgency3(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 0)
        #should be deadlock/timelock, committed location with no outgoing trans

    def test_urgent1(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x==0)')

        #should be deadlock, time not allowed to elapse
        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 0)

    def test_urgent_chan1(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x==0)')

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 1)

        s3 = succs2[0]
        succs3 = list(s.successors(s3))
        self.assertEqual(len(succs3), 0)

    def test_urgent_chan_choice(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 2)
        s2 = succs[1]
        #s2 is sync taken
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x==0)')

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 1)

        s3 = succs2[0]
        succs3 = list(s.successors(s3))
        self.assertEqual(len(succs3), 0)

        #s2 is sync not taken
        s2 = succs[0]
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x==0)')

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 0)

    def test_urgent_chan2(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), 'true')

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 1)
        s3 = succs2[0]
        s3fed = s3._get_pyfed_from_state()
        self.assertEqual(str(s3fed), '(x<=42)')
        succs3 = list(s.successors(s3))
        self.assertEqual(len(succs3), 0)

    def test_urgent_chan_guard(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 0)

    def test_urgent_chan_receiver_guard(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 0)

    def test_urgent_to_unurgent(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x<=42)')

        #should be deadlock, time not allowed to elapse
        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 0)

        #again, check that initial fed is still the same
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

    def test_unurgent_to_urgent(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(5<=x)')

        #should be deadlock
        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 0)

        #again, check that initial fed is still the same
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

    def test_constant_overrides(self):
        file = open(os.path.join(os.path.dirname(__file__), 'master-election-simple2.xml'))
        self.nta = pyuppaal.NTA.from_xml(file)
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 2)

        #override N
        s = GenerateSuccessorGenerator(self.nta, {'N': '4'})
        s1 = s.get_initialstate()

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 4)

    def test_traingate_1train(self):
        file = open(os.path.join(os.path.dirname(__file__), 'train-gate.xml'))
        self.nta = pyuppaal.NTA.from_xml(file)
        s = GenerateSuccessorGenerator(self.nta,  {'N': '1'}, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        #train0 approach
        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        self.assertEqual(s2.Train_0, 2)
        self.assertEqual(s2.Gate, 0)
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(Train_0_x<=20)')

        #train0 cross
        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 1)
        s3 = succs2[0]
        self.assertEqual(s3.Gate_len, 1)
        print s3
        #assert about dbm
        s3fed = s3._get_pyfed_from_state()
        self.assertEqual(str(s3fed), '(Train_0_x<=5)')

        #train0 safe
        succs3 = list(s.successors(s3))
        self.assertEqual(len(succs3), 1)
        s4 = succs3[0]
        print s4
        self.assertEqual(s4.Gate_len, 0)
        #assert about dbm
        s4fed = s4._get_pyfed_from_state()
        self.assertEqual(str(s4fed), 'true')

    def test_traingate_2train(self):
        file = open(os.path.join(os.path.dirname(__file__), 'train-gate.xml'))
        self.nta = pyuppaal.NTA.from_xml(file)
        s = GenerateSuccessorGenerator(self.nta,  {'N': '2'}, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 2)

        #train0 approach {{{
        s2 = succs[0]
        self.assertEqual(s2.Train_0, 2)
        self.assertEqual(s2.Gate, 0)
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(Train_0_x<=20)')
        #assert about array
        self.assertEqual(s2.Gate_len, 1)
        self.assertEqual(s2.Gate_list[0], 0)
        self.assertEqual(s2.Gate_list[1], 0)

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 2)
        #train0 cross {{{
        s3 = succs2[0]
        self.assertEqual(s3.Gate_len, 1)
        print s3
        #assert about dbm
        s3fed = s3._get_pyfed_from_state()
        self.assertEqual(str(s3fed), '(Train_0_x<=5)')

        #train0 safe
        succs3 = list(s.successors(s3))
        self.assertEqual(len(succs3), 2)
        s4 = succs3[0]
        print s4
        self.assertEqual(s4.Gate_len, 0)
        #assert about dbm
        s4fed = s4._get_pyfed_from_state()
        self.assertEqual(str(s4fed), 'true')
        #assert about array
        self.assertEqual(s4.Gate_len, 0)
        self.assertEqual(s4.Gate_list[0], 0)
        self.assertEqual(s4.Gate_list[1], 0)
        #}}}

        #other branch: train1 approach{{{
        s3 = succs2[1]
        self.assertEqual(s3.Gate_len, 2)
        self.assertEqual(s3.Gate, 2)
        print s3
        #assert about dbm
        s3fed = s3._get_pyfed_from_state()
        self.assertEqual(str(s3fed), '(Train_0_x<=20 & Train_1_x==0)')
        #assert about array
        self.assertEqual(s3.Gate_list[0], 0)
        self.assertEqual(s3.Gate_list[1], 1)

        #train1 stop
        succs3 = list(s.successors(s3))
        self.assertEqual(len(succs3), 1)
        s4 = succs3[0]
        print s4
        self.assertEqual(s4.Gate_len, 2)
        #assert about dbm
        s4fed = s4._get_pyfed_from_state()
        self.assertEqual(str(s4fed), '(Train_0_x<=20)')
        #}}}
        #}}}

    def test_dynamic_sync_sender(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        self.assertEqual(s1.i, 3)

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 0)

    def test_util_channame_index_exprs(self):
        self.assertEqual(
            util.channame_str_to_index_cexprs(""), 
            ("", [])
            )
        self.assertEqual(
            util.channame_str_to_index_cexprs("channame"), 
            ("channame", [])
            )
        self.assertEqual(
            util.channame_str_to_index_cexprs("channame[1]"), 
            ("channame[1]", [])
            )
        self.assertEqual(
            util.channame_str_to_index_cexprs("channame[1][2][3]"), 
            ("channame[1][2][3]", [])
            )
        self.assertEqual(
            util.channame_str_to_index_cexprs("channame[N]"), 
            ("channame", ["N"])
            )
        self.assertEqual(
            util.channame_str_to_index_cexprs("send[(0+1)%7][cur]"), 
            ("send", ["((0+1)%7)", "cur"])
            )

    def test_stopwatch1(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.P0, 1)
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]

        self.assertEqual(s2.P0, 0)
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x<=7 & x<=y & y<=8)')

    def test_stopwatch2(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.P0, 1)
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(y==0)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]

        self.assertEqual(s2.P0, 0)
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x<=7 & y<=x)')

    def test_urgent_bcast_chan1(self):
        s = GenerateSuccessorGenerator(self.nta, extrapolation="k")
        s1 = s.get_initialstate()
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(x==0)')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #assert about dbm
        s2fed = s2._get_pyfed_from_state()
        self.assertEqual(str(s2fed), '(x==0)')

        succs2 = list(s.successors(s2))
        self.assertEqual(len(succs2), 1)

        s3 = succs2[0]
        succs3 = list(s.successors(s3))
        self.assertEqual(len(succs3), 0)

    def test_upc_metamoc_cache(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        #assert empty cache
        for i in range(8):
            self.assertEqual(s1.Process_cache[i], 0)
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        #element in cache
        self.assertEqual(s2.Process_cache[0], 32)
        #assert rest empty cache
        for i in range(1, 8):
            self.assertEqual(s2.Process_cache[i], 0)

    def test_upc_functioncall(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        self.assertEqual(s1.x, 0)

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        self.assertEqual(s2.x, 42)

    def test_upc_functioncall_parameter(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        self.assertEqual(s1.x, 0)

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        self.assertEqual(s2.x, 42)

    def test_upc_functioncall_parameter_typedef(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        self.assertEqual(s1.x, 0)

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        self.assertEqual(s2.x, 42)
    
    def test_upc_functioncall_parameter_struct(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        print s1
        #self.assertEqual(s1.x, 0)

        succs = list(s.successors(s1))
        print succs
        #self.assertEqual(len(succs), 1)
        #s2 = succs[0]
        #self.assertEqual(s2.x, 42)

    def test_upc_functioncall_parameter_struct2(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        print s1
        #self.assertEqual(s1.x, 0)

        succs = list(s.successors(s1))

    def test_upc_func_struct_return(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        #print s1
        #self.assertEqual(s1.myStruct_t_x, 0)
        #self.assertEqual(s1.myStruct_t_y, 0)
        #succs = list(s.successors(s1))

        #self.assertEqual(len(succs), 1)
        #s2 = succs[0]
        #print s2
        #self.assertEqual(s2.myStruct_t_x, 3)
        #self.assertEqual(s2.myStruct_t_y, 100)

    def test_upc_array_init(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        self.assertEqual(s1.x[0], 0)
        self.assertEqual(s1.x[1], 2)
        #XXX: notice that the python access is currently flat 
        self.assertEqual(s1.y[1], 1)
        self.assertEqual(s1.y[2], 2)
        self.assertEqual(s1.y[5], 2)


    def test_upc_struct_init(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        self.assertEqual(s1.trivialint_a, 0)
        self.assertEqual(s1.trivialint_b, 0)
        self.assertEqual(s1.simpleint_a, 100)
        self.assertEqual(s1.simpleint_b, -190)
        self.assertEqual(s1.array_a[0], 0)
        self.assertEqual(s1.array_a[1], 9)
        self.assertEqual(s1.array_b[0], 2)
        self.assertEqual(s1.array_b[1], 3)
        #XXX: notice that the python access is currently flat 
        self.assertEqual(s1.nestedarray_a[0], -4)
        self.assertEqual(s1.nestedarray_a[1], -6)
        self.assertEqual(s1.nestedarray_a[2], 9)
        self.assertEqual(s1.nestedarray_a[3], 5)
        self.assertEqual(s1.nestedarray_b[0], 1)
        self.assertEqual(s1.nestedarray_b[1], 8)
        self.assertEqual(s1.nestedarray_b[2], 2)
        self.assertEqual(s1.nestedarray_b[3], 3)
        #self.assertEqual(s1.x, 0)

        succs = list(s.successors(s1))

    def test_guard_discrete_clock_mix(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        self.assertEqual(s1.x, 0)

        succs = list(s.successors(s1))
        #"x == 1" is NOT equal to "true"
        self.assertEqual(len(succs), 0)

    def test_invariant_var(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.x, 0)
        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 0)

    def test_invariant_mixed(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.x, 0)
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(c<=5)')

    def test_invariant_mixed_disjunct1(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.x, 0)
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), 'true')

    def test_invariant_mixed_disjunct2(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.x, 6)
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(c<=5)')

    def test_invariant_mixed_disjunct3(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.x, 6)
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(c<=5)')

    @unittest.skip("disabled until we have a type/syntax checker")
    def test_invariant_disallowed_disjunct(self):
        self.assertRaises(Exception, GenerateSuccessorGenerator, self.nta)

    def test_cfunc_wrong_ifcond(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.i, 1)

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        self.assertEqual(s2.i, 53)

    def test_signed_var(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.x, -1)

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        self.assertEqual(s2.x, -2)

    def test_upc_func_typedef_return(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()
        self.assertEqual(s1.x, 0)

        succs = list(s.successors(s1))
        self.assertEqual(len(succs), 1)
        s2 = succs[0]
        self.assertEqual(s2.x, 5)

    def test_invariant_mixed_disjunct4(self):
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        # initial state
        self.assertEqual(s1.x, 6)
        #assert about dbm
        initfed = s1._get_pyfed_from_state()
        self.assertEqual(str(initfed), '(c1<=6 & c1==c2)')

    def test_duplicate_functions(self):
        #check that it compiles
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

    def test_array_initializer1(self):
        #check that it compiles
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        self.assertEqual(s1.tstat_pri[0], 0)
        self.assertEqual(s1.tstat_status[0], 5)
        self.assertEqual(s1.tstat_pid[0], 7)
        self.assertEqual(s1.tstat_pri[1], 1)
        self.assertEqual(s1.tstat_status[1], 7)
        self.assertEqual(s1.tstat_pid[1], 5)

    @unittest.skip("initializer needs to parse general expressions")
    def test_array_initializer2(self):
        #check that it compiles
        s = GenerateSuccessorGenerator(self.nta)
        s1 = s.get_initialstate()

        self.assertEqual(s1.tstat[0].pri, 0)
        self.assertEqual(s1.tstat[0].status, 5)
        self.assertEqual(s1.tstat[0].pid, 7)
        self.assertEqual(s1.tstat[1].pri, 1)
        self.assertEqual(s1.tstat[1].status, 7)
        self.assertEqual(s1.tstat[1].pid, 5)

if __name__ == '__main__':
    verbose_level = sys.argv.count('-v')
    logging.basicConfig( stream=sys.stderr, 
            level={0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbose_level] )

    unittest.main() 

#vim ts=4:sw=4:expandtab
