#!/usr/bin/python
import sys
import os
import unittest
import logging

import pyuppaal
from opaal.model_parsers.pyuppaal.analyzer import AnalyzeModel
from opaal.model_parsers.pyuppaal.simplify import SimplifyModel

from opaal.model_parsers.pyuppaal import SuccessorGenerator, GoalChecker
from opaal.passed_waiting_list import SimplePWList
from opaal.modelchecking_algorithms.reachability import McReachability

class TestPYUPPAALAnalyzer(unittest.TestCase):

    def test_maxconst1(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst1.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        b = P.locations.index(P.get_location_by_name('b'))
        c = P.locations.index(P.get_location_by_name('c'))
        d = P.locations.index(P.get_location_by_name('d'))
        
        self.assertEqual(analyzer.max_constants[0][c]['y'], 100)
        self.assertEqual(analyzer.max_constants[0][c]['x'], -1)

        self.assertEqual(analyzer.max_constants[0][d]['y'], -1)
        self.assertEqual(analyzer.max_constants[0][d]['x'], -1)

        self.assertEqual(analyzer.max_constants[0][b]['y'], 100)
        self.assertEqual(analyzer.max_constants[0][b]['x'], 10)

    def test_maxconst2(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst2.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        b = P.locations.index(P.get_location_by_name('b'))

        self.assertEqual(analyzer.max_constants[0][a]['x'], 42)
        self.assertEqual(analyzer.max_constants[0][b]['x'], -1)

    def test_maxconst3(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst3.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        P_idx = 0
        a = P.locations.index(P.get_location_by_name('a'))
        b = P.locations.index(P.get_location_by_name('b'))
        c = P.locations.index(P.get_location_by_name('c'))
        d = P.locations.index(P.get_location_by_name('d'))

        self.assertEqual(analyzer.max_constants[0][a]['x'], 10)
        self.assertEqual(analyzer.max_constants[0][a]['y'], 42)
        for l in [b, c]:
            self.assertEqual(analyzer.max_constants[0][l]['x'], -1)
            self.assertEqual(analyzer.max_constants[0][l]['y'], 42)
        self.assertEqual(analyzer.max_constants[0][d]['x'], -1)
        self.assertEqual(analyzer.max_constants[0][d]['y'], -1)

        #test get_max_constants_for_state
        succgen = SuccessorGenerator(nta)
        succgen.is_variable('i')
        init_state = succgen.get_initialstate()

        init_state_maxconsts = analyzer.get_max_constants_for_state(init_state)
        self.assertEqual(init_state_maxconsts['x'], 10)
        self.assertEqual(init_state_maxconsts['y'], 42)

        #do delay until we get to next location
        succs = list(succgen.successors(init_state))
        while succs[0].locs[0] != b:
            self.assertEqual(len(succs), 1)
            succs = list(succgen.successors(succs[0]))

        self.assertEqual(succs[0].locs[0], b)

        succ_maxconsts = analyzer.get_max_constants_for_state(succs[0])
        self.assertEqual(succ_maxconsts['x'], -1)
        self.assertEqual(succ_maxconsts['y'], 42)



    def test_maxconst4(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst4.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        succgen = SuccessorGenerator(nta)
        
        #this will fail, unless max constant correctly detected
        goalchecker = GoalChecker(succgen, 'E<> P.c')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertTrue(res)


    def test_viking1(self):
        file = open(os.path.join(os.path.dirname(__file__), 'vikingtests', 'viking1.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        Viking1 = nta.templates[0]
        Viking1_idx = 0
        unsafe_idx = Viking1.locations.index(Viking1.get_location_by_name('unsafe'))
        safe_idx = Viking1.locations.index(Viking1.get_location_by_name('safe'))

        #this fails unless clock resets are taken into account
        self.assertEqual(analyzer.max_constants[Viking1_idx][unsafe_idx]['y'], -1)
        self.assertEqual(analyzer.max_constants[Viking1_idx][safe_idx]['y'], -1)

        for (l_idx, l) in enumerate(Viking1.locations):
            if not l.name.value in ('unsafe', 'safe'):
                self.assertEqual(analyzer.max_constants[Viking1_idx][l_idx]['y'], 5)

    def test_lubounds_example(self):
        file = open(os.path.join(os.path.dirname(__file__), 'lubounds_example.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        t = nta.templates[0]
        t_idx = 0

        l0_idx = t.locations.index(t.get_location_by_name('l0'))
        l1_idx = t.locations.index(t.get_location_by_name('l1'))
        l2_idx = t.locations.index(t.get_location_by_name('l2'))

        #max constants
        #l1
        self.assertEqual(analyzer.max_constants[t_idx][l1_idx]['x'], -1)
        self.assertEqual(analyzer.max_constants[t_idx][l1_idx]['y'], -1)

        #l2
        self.assertEqual(analyzer.max_constants[t_idx][l2_idx]['x'], -1)
        self.assertEqual(analyzer.max_constants[t_idx][l2_idx]['y'], -1)

        #l0
        self.assertEqual(analyzer.max_constants[t_idx][l0_idx]['x'], 1000000)
        self.assertEqual(analyzer.max_constants[t_idx][l0_idx]['y'], 1)

        #lower-upper bounds
        #l1
        self.assertEqual(analyzer.upperbound[t_idx][l1_idx]['x'], -1)
        self.assertEqual(analyzer.upperbound[t_idx][l1_idx]['y'], -1)
        self.assertEqual(analyzer.lowerbound[t_idx][l1_idx]['x'], -1)
        self.assertEqual(analyzer.lowerbound[t_idx][l1_idx]['y'], -1)

        #l2
        self.assertEqual(analyzer.upperbound[t_idx][l2_idx]['x'], -1)
        self.assertEqual(analyzer.upperbound[t_idx][l2_idx]['y'], -1)
        self.assertEqual(analyzer.lowerbound[t_idx][l2_idx]['x'], -1)
        self.assertEqual(analyzer.lowerbound[t_idx][l2_idx]['y'], -1)

        #l0
        self.assertEqual(analyzer.upperbound[t_idx][l0_idx]['x'], 1000000)
        self.assertEqual(analyzer.upperbound[t_idx][l0_idx]['y'], 1)
        self.assertEqual(analyzer.lowerbound[t_idx][l0_idx]['x'], 10)
        self.assertEqual(analyzer.lowerbound[t_idx][l0_idx]['y'], 1)


    def test_maxconst_constvar(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_constvar.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        self.assertEqual(analyzer.max_constants[0][a]['c'], 5)

    def test_maxconst_constvar2(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_constvar2.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        self.assertEqual(analyzer.max_constants[0][a]['c'], 5)

    def test_manyclocks(self):
        file = open(os.path.join(os.path.dirname(__file__), 'ltl_tests', 'hddi_input_10.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        simple_nta = SimplifyModel(nta).simplify()

        self.assertEqual(len(simple_nta.templates), 11)

        analyzer = AnalyzeModel(simple_nta)
        analyzer.analyze()

        for i in range(1, 11):
            STx = nta.templates[i]
            self.assertEqual(STx.name, "ST%d" % i)
            a = STx.locations.index(STx.get_location_by_name('station_z_idle'))
        
            self.assertEqual(analyzer.max_constants[i][a]['ST%d_z' % i], 520)

    def test_maxconst_var1(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_var1.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        #UPPAAL int by default has range [-32767, 32767]
        self.assertEqual(analyzer.max_constants[0][a]['c'], 32767)

    def test_maxconst_var2(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_var2.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        #int defined with range [0, 4]
        self.assertEqual(analyzer.max_constants[0][a]['c'], 4)

    def test_maxconst_expr1(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_expr1.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        #[0,4] + [1,1] = [1,5]
        self.assertEqual(analyzer.max_constants[0][a]['c'], 5)

    def test_maxconst_expr2(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_expr2.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        #-5 + 14
        self.assertEqual(analyzer.max_constants[0][a]['c'], 9)

    def test_maxconst_expr3(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_expr3.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        self.assertEqual(analyzer.max_constants[0][a]['c'], 8)

    def test_maxconst_expr4(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_expr4.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        self.assertEqual(analyzer.max_constants[0][a]['c'], 4)

    def test_maxconst_expr5(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_expr5.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        self.assertEqual(analyzer.max_constants[0][a]['c'], 8)

    def test_maxconst_expr6(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_expr6.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        self.assertEqual(analyzer.max_constants[0][a]['c'], -1)

    def test_maxconst_expr7(self):
        file = open(os.path.join(os.path.dirname(__file__), 'maxconst_expr7.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        analyzer = AnalyzeModel(nta)
        analyzer.analyze()

        P = nta.templates[0]
        a = P.locations.index(P.get_location_by_name('a'))
        
        self.assertEqual(analyzer.max_constants[0][a]['c'], 5500)
        #better analysis could find bound to be 170

if __name__ == '__main__':
    verbose_level = sys.argv.count('-v')
    logging.basicConfig( stream=sys.stderr, 
            level={0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbose_level] )

    unittest.main() 
