#!/usr/bin/python
import sys
import os
import unittest
import logging
import copy

import pyuppaal
from opaal.model_parsers.pyuppaal import SuccessorGenerator
from opaal.lattices import DBMClock, DBMFederation


class TestPYUPPAALLattice(unittest.TestCase):
    def test_clock_typedef(self):
        file = open(os.path.join(os.path.dirname(__file__), 'dbmtests', 'simpleclocktrans.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        
        s = SuccessorGenerator(nta)
        self.assertEqual(s.variables, [
            ('dbm', 'DBMFederation', [], None),
            ('x', 'DBMClock', [], None),
            ('c', 'DBMClock', [], None),
            ])

        initial_state = s.get_initialstate()
        self.assertEqual(initial_state['dbm'].__class__, DBMFederation)

        self.assertEqual(str(initial_state['dbm']), '(dbm.x==dbm.c)')
        
        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 2)

        #XXX TODO MORE ASSERTS

    def test_dbmlatticeclass(self):
        dbmfed = DBMFederation(name='dbmfed', dimensions=None, context=self)
        self.assertEqual(dbmfed.context.clocks, [])

        a = DBMClock(name='a', context=self)
        self.assertEqual(len(dbmfed.context.clocks), 1)
        self.assertTrue(dbmfed.context.a)
        b = DBMClock(name='b', context=self)
        self.assertEqual(len(dbmfed.context.clocks), 2)
        self.assertTrue(dbmfed.context.b)

        self.assertEqual(str(dbmfed), '(dbmfed.a==dbmfed.b & dbmfed.b==0)')
        dbmfed = dbmfed.up()
        self.assertEqual(str(dbmfed), '(dbmfed.a==dbmfed.b)')


    def test_dbmfed_copy(self):
        dbmfed = DBMFederation(name='dbmfed', dimensions=None, context=self)
        a = DBMClock(name='a', context=self)
        b = DBMClock(name='b', context=self)

        dbmfed2 = copy.copy(dbmfed)
        self.assertEqual(dbmfed, dbmfed2)

        dbmfed2 = copy.deepcopy(dbmfed)
        self.assertEqual(dbmfed, dbmfed2)

    def test_dbmfed_ops(self):
        dbmfed = DBMFederation(name='dbmfed', dimensions=None, context=self)
        a = DBMClock(name='a', context=self)
        b = DBMClock(name='b', context=self)

        f = dbmfed.copy()
        self.assertTrue(f.up().covers(f))
        f = f.up()
        self.assertTrue(f.covers(f & (f.a <= 3)))
        self.assertTrue((f & (f.a <= 4)).covers(f & (f.a <= 3)))
        #import pdb; pdb.set_trace()
        self.assertFalse((f & (f.a <= 2)).covers(f & (f.a <= 3)))


    def test_invariant_delay(self):
        file = open(os.path.join(os.path.dirname(__file__), 'dbmtests', 'simpleinvariantdelay.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        
        s = SuccessorGenerator(nta)
        self.assertEqual(s.variables, [
            ('dbm', 'DBMFederation', [], None),
            ('x', 'DBMClock', [], None),
            ('c', 'DBMClock', [], None),
            ])

        initial_state = s.get_initialstate()

        self.assertEqual(str(initial_state['dbm']), '(dbm.x<=4 & dbm.x==dbm.c)')
        
        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 1)

        succ = succs[0]
        self.assertEqual(str(succ['dbm']), '(dbm.x==4 & dbm.x==dbm.c & dbm.c==4)')

        succs = list(s.successors(succ))
        self.assertEqual(len(succs), 0)




if __name__ == '__main__':
    verbose_level = sys.argv.count('-v')
    logging.basicConfig( stream=sys.stderr, 
            level={0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbose_level] )

    unittest.main() 

# vim:ts=4:sw=4:expandtab
