#!/usr/bin/python
import sys
import os
import unittest
import logging

import pyuppaal
from opaal.model_parsers.pyuppaal import SuccessorGenerator, GoalChecker
from opaal.passed_waiting_list import SimplePWList
from opaal.modelchecking_algorithms.reachability import McReachability
from opaal.model_parsers.pyuppaal.simplify import SimplifyModel
from opaal.model_parsers.pyuppaal import util

class TestPYUPPAALSuccGen(unittest.TestCase):

    def test_invariant(self):
        file = open(os.path.join(os.path.dirname(__file__), 'test_invariant.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        
        s = SuccessorGenerator(nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 0)


    def test_simplify(self):
        file = open(os.path.join(os.path.dirname(__file__), 'master-election-simple.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        simple_nta = SimplifyModel(nta).simplify()
        
        #Test that templates have been expanded
        self.assertEqual(len(simple_nta.templates), 7)

        #test select has been expanded
        temp0 = [t for t in simple_nta.templates if t.name == 'Node_0'][0]
        self.assertEqual(len(temp0.transitions), 9)

    def test_simple_sync(self):
        file = open(os.path.join(os.path.dirname(__file__), 'test_simple_sync.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        
        s = SuccessorGenerator(nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 1)
        
        succs2 = list(s.successors(succs[0]))
        self.assertEqual(len(succs2), 0)

    def test_sync_twoway(self):
        file = open(os.path.join(os.path.dirname(__file__), 'sync_twoway.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        
        s = SuccessorGenerator(nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 1)
        
        succs2 = list(s.successors(succs[0]))
        self.assertEqual(len(succs2), 1)

        succs3 = list(s.successors(succs2[0]))
        self.assertEqual(len(succs3), 0)

    def test_sync_chan_expression(self):
        file = open(os.path.join(os.path.dirname(__file__), 'sync_chan_expression.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        
        s = SuccessorGenerator(nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 1)
        
        succs2 = list(s.successors(succs[0]))
        self.assertEqual(len(succs2), 1)

        succs3 = list(s.successors(succs2[0]))
        self.assertEqual(len(succs3), 0)

    def test_sync_bcast(self):
        file = open(os.path.join(os.path.dirname(__file__), 'sync_bcast.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        
        s = SuccessorGenerator(nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        #B can do independent, or A can do action
        self.assertEqual(len(succs), 2)

        #early sync branch
        earlysync_succ = [succ for succ in succs if succ.locs[0] == 1][0]
        succs2 = list(s.successors(earlysync_succ))
        self.assertEqual(len(succs2), 1)
        succs3 = list(s.successors(succs2[0]))
        self.assertEqual(len(succs3), 0)
        #B has not synced
        self.assertEqual(1, succs2[0].locs[1])

        #late sync branch
        latesync_succ = [succ for succ in succs if succ.locs[0] == 0][0]
        succs2 = list(s.successors(latesync_succ))
        self.assertEqual(len(succs2), 1)
        succs3 = list(s.successors(succs2[0]))
        self.assertEqual(len(succs3), 0)
        #B has synced
        self.assertEqual(2, succs2[0].locs[1])


    def test_sync_bcast_many(self):
        file = open(os.path.join(os.path.dirname(__file__), 'sync_bcast_many.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        succgen = SuccessorGenerator(nta)
        
        #check that all participate
        goalchecker = GoalChecker(succgen, 'E<> mysum != 0 and mysum != 15')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertFalse(res)

    def test_array_init(self):
        file = open(os.path.join(os.path.dirname(__file__), 'test_array_init.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        
        s = SuccessorGenerator(nta)

        self.assertEqual(s.variables, [('test', 'TypeInt', [2, 3], 0), ('test2', 'TypeInt', [2, 3, 4], 0)])

        initial_state = s.get_initialstate()
        self.assertEqual(initial_state['test'], [[0, 0, 0], [0, 0, 0]])
        self.assertEqual(initial_state['test2'], [[[0]*4]*3]*2)

        initial_state['test'][0][0] = 42
        self.assertEqual(initial_state['test'], [[42, 0, 0], [0, 0, 0]])

    def test_simplify_channel_array(self):
        file = open(os.path.join(os.path.dirname(__file__), 'test_channel_array.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        simple_nta = SimplifyModel(nta).simplify()
        #test select has been expanded
        temp0 = [t for t in simple_nta.templates if t.name == 'Template_0'][0]
        self.assertEqual(len(temp0.transitions), 1)
        trans0 = temp0.transitions[0]
        self.assertEqual(trans0.synchronisation.value, 'mychan[Template_0_id]!')

        #test synchronisation on channel array
        s = SuccessorGenerator(simple_nta)
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        self.assertEqual(len(succs), 3)

    def test_util_expression_to_python(self):
        self.assertEqual(
            util.expression_str_to_python("dbm.isEmpty() == False"), 
            "dbm.isEmpty() == False"
            )
        self.assertEqual(
            util.expression_str_to_python("dbm.isEmpty()"), 
            "dbm.isEmpty()"
            )
        self.assertEqual(
            util.expression_str_to_python("(x == 4 && y)"), 
            "x == 4 and y"
            )
        self.assertEqual(
            util.expression_str_to_python("true"), 
            "True"
            )
        self.assertEqual(
            util.expression_str_to_python("True"), 
            "True"
            )
        self.assertEqual(
            util.expression_str_to_python("msg[0]"), 
            "msg[0]"
            )
        self.assertEqual(
            util.expression_str_to_python("msg[0][0] == true && 0 > N-0-1"), 
            "msg[0][0] == True and 0 > N - 0 - 1"
            )

    def test_util_expression_to_c(self):
        class StubSuccGen:
            variables = []
            discrete_variables = []
            lattice_variables = []
            local_variables = []
            clocks = []
            constants = []
            structs = {}
            simplyfied_struct_name = {}
            class myparser:
                identifierTypeDict = {}
            pars = myparser()

            def __init__(self):
                self.variables.append(("x", "int", 0, 0))
                self.variables.append(("y", "int", 0, 0))
                self.variables.append(("msg", "int", 0, 0))
                self.variables.append(("N", "int", 0, 0))
                self.discrete_variables = self.variables

            def is_variable(self, identifier):
                if identifier in [name for (name, _, _, _) in self.variables]:
                    return True

                return False
            def is_lattice_variable(self, identifier):
                return False


        self.assertEqual(
            util.expression_str_to_c("dbm.isEmpty() == false"), 
            "(dbm.isEmpty() == false)"
            )
        self.assertEqual(
            util.expression_str_to_c("dbm.isEmpty()"), 
            "dbm.isEmpty()"
            )
        self.assertEqual(
            util.expression_str_to_c("(x == 4 && y)", StubSuccGen()), 
            "((in->x == 4)) && (in->y)"
            )
        self.assertEqual(
            util.expression_str_to_c("true"), 
            "true"
            )
        self.assertEqual(
            util.expression_str_to_c("false"), 
            "false"
            )
        self.assertEqual(
            util.expression_str_to_c("msg[0]", StubSuccGen()), 
            "in->msg[0]" 
            )
        self.assertEqual(
            util.expression_str_to_c("msg[0][0] == true && 0 > N-0-1", StubSuccGen()), 
            "((in->msg[0][0] == true)) && ((0 > ((in->N - 0) - 1)))"
            )
        self.assertEqual(
            util.expression_str_to_c("4 + 5"), 
            "(4+5)"
            )
        self.assertEqual(
            util.expression_str_to_c("4 % 5"), 
            "(4%5)"
            )
        self.assertEqual(
            util.expression_str_to_c("(4 % 5)"), 
            "(4%5)"
            )

        class StubSuccGen2:
            variables = []
            local_variables = []
            clocks = []
            constants = []
            structs = {}
            simplyfied_struct_name = {}
            class myparser:
                identifierTypeDict = {}
            pars = myparser()

            def __init__(self):
                self.clocks.append(("c", 10))
                self.variables.append(("x", "int", 0, 0))
                self.discrete_variables = self.variables

            def is_variable(self, identifier):
                if identifier in [name for (name, _, _, _) in self.variables]:
                    return True
                return False
            def is_lattice_variable(self, identifier):
                return False

        s2 = StubSuccGen2()

        self.assertEqual(
            util.expression_str_to_c("c > 5", s2), 
            "true"
            )
        self.assertEqual(
            util.expression_str_to_c("x > 5", s2), 
            "(in->x > 5)"
            )

    def test_util_statement_to_c(self):
        class ClockStubSuccGen:
            clocks = []
            clock_indexes = []
            variables = [("g", "int", 0, 0), ("id", "int", 0, 0), ("pid", "int", 0, 0), ("i", "int", 0, 0)]
            discrete_variables = variables
            local_variables = []
            structs = {}
            simplyfied_struct_name = {}
            class myparser:
                identifierTypeDict = {}
            pars = myparser()

            def __init__(self):
                self.clocks.append(("y", 0))
                self.clocks.append(("x", 0))
                self.clock_indexes = ["0clock"] + [c for (c, _) in self.clocks]

            def is_variable(self, identifier):
                if identifier in [name for (name, _, _, _) in self.variables]:
                    return True
                return False
            def is_lattice_variable(self, identifier):
                return False

        sg = ClockStubSuccGen()
        self.assertEqual(util.statement_str_to_c("i = g", sg).replace('\n', ''),"out->i = out->g;")
        self.assertEqual(util.statement_str_to_c("x = g", sg).replace('\n', ''),"out->fed->updateValue(2, out->g);")
        self.assertEqual(util.statement_str_to_c("g = 0", sg).replace('\n', ''),"out->g = 0;")
        self.assertEqual(util.statement_str_to_c("g = 0,\n id = pid", sg).replace('\n', ''),"out->g = 0;out->id = out->pid;")
        self.assertEqual(util.statement_str_to_c("x = 0,\n id = pid", sg).replace('\n', ''),"out->fed->updateValue(2, 0);out->id = out->pid;")
        self.assertEqual(util.statement_str_to_c("x = 0, id = pid, y = 0", sg).replace('\n', ''),"out->fed->updateValue(2, 0);out->id = out->pid;out->fed->updateValue(1, 0);")


    def test_util_expression_to_fedupdate_c(self):
        class StubSuccGen:
            variables = []
            clocks = []
            clock_indexes = []
            constants = {}
            structs = {}
            simplyfied_struct_name = {}
            class myparser:
                identifierTypeDict = {}
            pars = myparser()

            def __init__(self):
                self.clocks.append(("c", 10))
                self.clock_indexes.append("0clock")
                self.clock_indexes.append("c")
                self.variables.append(("x", "int", 0, 0))
                self.discrete_variables = self.variables

            def is_variable(self, identifier):
                if identifier in [name for (name, _, _, _) in self.variables]:
                    return True

                return False
            def is_lattice_variable(self, identifier):
                return False

        s = StubSuccGen()

        self.assertEqual(
            util.expression_str_to_fedupdate_c("c > 5", s), 
            "*(out->fed) &= constraint_t(0, 1, -5, true); "
            )
        self.assertEqual(
            util.expression_str_to_fedupdate_c("x > 5", s), 
            ""
            )
        self.assertEqual(
            util.expression_str_to_fedupdate_c("c > x", s), 
            "*(out->fed) &= constraint_t(0, 1, -out->x, true); "
            )


    def test_sup_query(self):
        file = open(os.path.join(os.path.dirname(__file__), 'suptest.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        succgen = SuccessorGenerator(nta)
        
        #check that all participate
        goalchecker = GoalChecker(succgen, 'sup: x')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        supval = goalchecker.finalize()

        self.assertFalse(res)
        self.assertEqual(supval, 42)

    def test_sync_self(self):
        file = open(os.path.join(os.path.dirname(__file__), 'sync_self.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        s = SuccessorGenerator(nta)
        
        initial_state = s.get_initialstate()

        succs = list(s.successors(initial_state))
        #sync with self should not be allowed
        self.assertEqual(len(succs), 0)

    def test_sync_bcast_self(self):
        file = open(os.path.join(os.path.dirname(__file__), 'sync_bcast_self.xml'))
        nta = pyuppaal.NTA.from_xml(file)
        succgen = SuccessorGenerator(nta)

        goalchecker = GoalChecker(succgen, 'E<> P.b')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()

        self.assertFalse(res)

        goalchecker = GoalChecker(succgen, 'E<> P.a')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()

        self.assertTrue(res)

    def test_util_channame_to_python_format_string(self):
        self.assertEqual(
            util.channame_str_to_python_format_string(""), 
            (True, "")
            )
        self.assertEqual(
            util.channame_str_to_python_format_string("channame"), 
            (True, "channame")
            )
        self.assertEqual(
            util.channame_str_to_python_format_string("channame[1]"), 
            (True, "channame[1]")
            )
        self.assertEqual(
            util.channame_str_to_python_format_string("channame[N]"), 
            (False, "channame[%(N)s]")
            )
        self.assertEqual(
            util.channame_str_to_python_format_string("send[(0+1)%7][cur]"), 
            (False, "send[1][%(cur)s]")
            )
    
    def test_simplify_expand_system_declaration_simple(self):
        #this is an empty UPPAAL model, just saved
        #it uses old-style system declarations
        file = open(os.path.join(os.path.dirname(__file__), 'empty-model.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        simple_nta = SimplifyModel(nta).simplify()
        
        #Test that templates have been expanded
        self.assertEqual(len(simple_nta.templates), 1)

    def test_simplify_expand_system_declaration_advanced(self):
        file = open(os.path.join(os.path.dirname(__file__), 'SimpleLeader3_mod.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        simple_nta = SimplifyModel(nta).simplify()
        
        #Test that templates have been expanded
        self.assertEqual(len(simple_nta.templates), 6)

        succgen = SuccessorGenerator(simple_nta)

        goalchecker = GoalChecker(succgen, 'E<> N2.Leader')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        self.assertEqual(init_state['N0_cur'], 1)
        pwlist.addState(init_state)
        res = m.start()
        self.assertTrue(res)

        goalchecker = GoalChecker(succgen, 'E<> N1.Leader')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertFalse(res)

    def test_typedef_var(self):
        file = open(os.path.join(os.path.dirname(__file__), 'typedef_var.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        self.assertEqual(len(nta.templates), 1)

        succgen = SuccessorGenerator(nta)
        init_state = succgen.get_initialstate()

        self.assertEqual(init_state['myvar'], 1)

        succs = list(succgen.successors(init_state))
        #sync with self should not be allowed
        self.assertEqual(len(succs), 1)
        succ = succs[0]

        self.assertEqual(succ['myvar'], 2)

    def test_simpleleader3(self):
        file = open(os.path.join(os.path.dirname(__file__), 'SimpleLeader3_dbm.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        succgen = SuccessorGenerator(nta)

        goalchecker = GoalChecker(succgen, 'E<> N2.Leader')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        self.assertEqual(init_state['N0_cur'], 1)
        pwlist.addState(init_state)
        res = m.start()
        self.assertTrue(res)

        goalchecker = GoalChecker(succgen, 'E<> N1.Leader')
        pwlist = SimplePWList()
        m = McReachability(succgen, goalchecker, pwlist)
        init_state = succgen.get_initialstate()
        pwlist.addState(init_state)
        res = m.start()
        self.assertFalse(res)

    def test_simplify_tapaal(self):
        file = open(os.path.join(os.path.dirname(__file__), 'tapaal', 'train-crossing-optstdred-5.xml'))
        nta = pyuppaal.NTA.from_xml(file)

        simple_nta = SimplifyModel(nta).simplify()

        self.assertTrue(True)
        

if __name__ == '__main__':
    verbose_level = sys.argv.count('-v')
    logging.basicConfig( stream=sys.stderr, 
            level={0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}[verbose_level] )

    unittest.main() 
